
//Ethan Alexander Shulman 2020 - xaloez.com

//html dom elements
var fileSelectEle,statusText,bckButton,nxtButton,modeSelect,
bgCanvas,bg2d,

//arrays of images
styleImages = null,transferImages = null,imageResults = [],

//webgl graphics objects
styleTextures = null, styleRT = null,
shaderPrograms = new Array(10), fbuf,
synthTex = null, transferTextures = null,

//generator state and options
styleId,imgId,imgX,imgY,
patternSz,quantization,synthWidth,synthHeight,
countBuffer,countBytes,simBuffer,simMinCount,simMaxCount,
synth2D,synthData,synthSourceData,
mode,imageTiling,sourceTiling,imageSeamPasses = 0,nSeamPasses = 0,
startTime,
stepId = -1;

//update displayed images, clear display html and add if neccesary
function updateDisplay() {
	var ds = document.getElementById("dis");
	ds.innerHTML = "";
	if (stepId === 0 || (stepId === 1 && modeSelect.selectedIndex === 1) || stepId >= 2) {
		var si = stepId===1?transferImages:(stepId>=2?imageResults:styleImages);
		if (si) {
			for (var i = 0; i < si.length; i++) ds.appendChild(si[i]);
		}
	}
}

//check if current step is completed, e.g if images have been selected
function stepComplete() {
	if (stepId === 0) {
		return styleImages && styleImages.length;
	} else if (stepId === 1 && modeSelect.selectedIndex === 1) {
		return transferImages && transferImages.length;
	}
	return true;
}

function timeStr(ms) {
	var pdate = new Date(ms),
		seconds = pdate.getUTCSeconds(), minutes = pdate.getUTCMinutes(), hours = pdate.getUTCHours();
	return (hours?hours+"h, ":"")+(minutes?minutes+"m, ":"")+seconds+"s.";
}

//main processing loop
function gen() {
	if (stepId !== 2) {
		//processing has been cancelled
		UnbindTexture(currentShader["img"]);
		return;
	}

	var lastTimeout = Date.now(),
		prog = 0;
	if (mode === 0) {
		if (nSeamPasses) prog = (imageSeamPasses-1+imgY/synthHeight)*100/nSeamPasses;
		else prog = imgY*100/synthHeight;
	} else if (mode === 1) {
		prog = (imgId+imgY/transferImages[imgId].height)*100/transferImages.length;
	} else if (mode === 2) {
		prog = (imgId+imgY/styleImages[imgId].height)*100/styleImages.length;
	}
	statusText.innerHTML = "Step 3:</br>Generating "+prog.toFixed(2)+"% complete.</br>"+
						"Estimated time left: "+timeStr(((lastTimeout-startTime)/prog)*(100-prog));
						
	while (Date.now()-lastTimeout < 100) {
		if (stepId !== 2) {
			//processing has been cancelled
			UnbindTexture(currentShader["img"]);
			return;
		}
	
		BindTexture(currentShader["src"],styleTextures[styleId]);
		var simg = styleImages[styleId];
		fbuf[0] = simg.width;
		fbuf[1] = simg.height;
		gl.uniform4fv(currentShader["data"],fbuf);
		styleRT[styleId].Bind();
		gl.drawArrays(gl.TRIANGLES,0,3);
		UnbindTexture(currentShader["src"]);
	
		styleId++;
		var nStyle = styleImages.length;
		if (styleId >= nStyle) {
			//done rendering style similarities, count up results
			if (mode === 0) {//synthesize image
				
				var maxC = 1, found = [];
				for (var si = 0; si < nStyle; si++)	{
					simg = styleImages[si];
					var swidth = simg.width, sheight = simg.height;
					styleRT[si].Bind();
					gl.readPixels(0,0,swidth,sheight,gl.RGBA,gl.UNSIGNED_BYTE,countBytes);
				
					var ci = 0, thid = si*countBuffer.length;
					for (var y = 0; y < sheight; y++) {
						for (var x = 0; x < swidth; x++) {
							var v = countBuffer[ci++];
							if (v > maxC) {
								maxC = v;
								found.length = 0;
								found.push(thid);
							} else if (v === maxC) {
								found.push(thid);
							}
							thid++;
						}
					}
				}
			
				var pxb = synthData.data,sid,pid;
				if (found.length) {
					//randomly select pixel for synthesized image from found
					pid = found[Math.floor(Math.random()*found.length)];
					sid = Math.floor(pid/countBuffer.length);
					pid = (pid-sid*countBuffer.length)*4;
				} else {
					//if non found randomly insert pixel from sources
					sid = Math.floor(Math.random()*nStyle);
					pid = Math.floor(Math.random()*synthSourceData[sid].data.length/4)*4;
				}
				var spx = synthSourceData[sid].data;
				pxb[0] = spx[pid++];
				pxb[1] = spx[pid++];
				pxb[2] = spx[pid++];
				pxb[3] = spx[pid++];
				synth2D.putImageData(synthData,imgX,imgY);
				gl.bindTexture(gl.TEXTURE_2D,synthTex);
				gl.texSubImage2D(gl.TEXTURE_2D,0,imgX,imgY,1,1,gl.RGBA,gl.UNSIGNED_BYTE,synthData.data);
			
				styleId = 0;
				imgX++;
				fbuf[2] = imgX;
				if (imgX >= synthWidth) {
					imgX = 0;
					fbuf[2] = 0;
					imgY++;
					fbuf[3] = imgY;
					if (imgY >= synthHeight) {
						UnbindTexture(currentShader["img"]);
						var finished = false;
						if (imageTiling) {
							if (imageSeamPasses < nSeamPasses) {
								imgX = 0;
								imgY = 0;
								fbuf[3] = 0;
								shaderPrograms[sourceTiling?9:8].Use();
								BindTexture(currentShader["img"],synthTex);
								imageSeamPasses++;
							} else {
								finished = true;
							}
						} else {
							finished = true;
						}
						if (finished) {
							//convert generated canvas element into image
							var ce = imageResults[0],
								img = document.createElement("img");
							img.src = ce.toDataURL();
							imageResults[0] = img;
						
							loadStep(stepId+1);
							return;
						}
					}
				}
			
			} else if (mode === 1) {//transform existing mode
			
				var maxC = 1, found = [];
				for (var si = 0; si < nStyle; si++)	{
					simg = styleImages[si];
					var swidth = simg.width, sheight = simg.height;
					styleRT[si].Bind();
					gl.readPixels(0,0,swidth,sheight,gl.RGBA,gl.UNSIGNED_BYTE,countBytes);
				
					var ci = 0, thid = si*countBuffer.length;
					for (var y = 0; y < sheight; y++) {
						for (var x = 0; x < swidth; x++) {
							var v = countBuffer[ci++];
							if (v > maxC) {
								maxC = v;
								found.length = 0;
								found.push(thid);
							} else if (v === maxC) {
								found.push(thid);
							}
							thid++;
						}
					}
				}
			
				var pxb = synthData.data,sid,pid;
				if (found.length) {
					//randomly select pixel for synthesized image from found
					pid = found[Math.floor(Math.random()*found.length)];
					sid = Math.floor(pid/countBuffer.length);
					pid = (pid-sid*countBuffer.length)*4;
				} else {
					//if non found randomly insert pixel from sources
					sid = Math.floor(Math.random()*nStyle);
					pid = Math.floor(Math.random()*synthSourceData[sid].data.length/4)*4;
				}
				var spx = synthSourceData[sid].data;
				pxb[0] = spx[pid++];
				pxb[1] = spx[pid++];
				pxb[2] = spx[pid++];
				pxb[3] = spx[pid++];
				synth2D.putImageData(synthData,imgX,imgY);

				styleId = 0;
				imgX++;
				fbuf[2] = imgX;
				if (imgX >= synthWidth) {
					imgX = 0;
					fbuf[2] = 0;
					imgY++;
					fbuf[3] = imgY;
					if (imgY >= synthHeight) {
					
						//convert generated canvas element into image
						var ce = imageResults[imgId],
							img = document.createElement("img");
						img.src = ce.toDataURL();
						imageResults[imgId] = img;
						
						UnbindTexture(currentShader["img"]);
						
						imgId++;
						if (imgId >= transferImages.length) {
							BindTexture(currentShader["img"],transferTextures[imgId]);
							loadStep(stepId+1);
							return;
						} else {
							
							var transfer = transferImages[imgId];
							fbuf[4] = synthWidth = ce.width = transfer.width;
							fbuf[5] = synthHeight = ce.height = transfer.height;
							imageResults.push(ce);
							updateDisplay();
							
							fbuf[3] = 0;
						}
					}
				}
			
			} else if (mode === 2) {//visualize similarity
			
				var maxC = 1,maxRep = 0;
				for (var si = 0; si < nStyle; si++)	{
					simg = styleImages[si];
					var swidth = simg.width, sheight = simg.height;
					styleRT[si].Bind();
					gl.readPixels(0,0,swidth,sheight,gl.RGBA,gl.UNSIGNED_BYTE,new Uint8Array(countBuffer.buffer));
				
					var ci = 0;
					for (var y = 0; y < sheight; y++) {
						for (var x = 0; x < swidth; x++) {
							var v = countBuffer[ci++];
							if (v > maxC) {
								maxRep = 1;
								maxC = v;
							} else if (v === maxC) maxRep++;
						}
					}
				}
				maxC *= maxRep;
				simMinCount = Math.min(simMinCount,maxC);
				simMaxCount = Math.max(simMaxCount,maxC);
			
				var cimg = styleImages[imgId];
				simBuffer[imgX+imgY*cimg.width] = maxC;
			
				styleId = 0;
				imgX++;
				fbuf[2] = imgX;
				if (imgX >= cimg.width) {
					imgX = 0;
					fbuf[2] = 0;
					imgY++;
					fbuf[3] = imgY;
					if (imgY >= cimg.height) {
						//output normalized similarity values to canvas in heatmap coloring
						var iwidth = cimg.width, iheight = cimg.height;
						bgCanvas.width = iwidth;
						bgCanvas.height = iheight;
						var idat = bg2d.getImageData(0,0,iwidth,iheight),
							px = idat.data, pi = 0, si = 0,
							iscale = 1/Math.max(1e-20,simMaxCount-simMinCount);
						for (var y = 0; y < iheight; y++) {
							for (var x = 0; x < iwidth; x++) {
								var v = Math.floor(Math.pow((simBuffer[si++]-simMinCount)*iscale,0.5)*255);
								px[pi++] = v;
								px[pi++] = Math.max(0,v-127)*2;
								px[pi++] = 255-v;
								px[pi++] = 255;
							}
						}
						bg2d.putImageData(idat,0,0);
					
						var img = document.createElement("img");
						img.src = bgCanvas.toDataURL();
						imageResults.push(img);
						updateDisplay();
			
						imgId++;
						if (imgId >= styleImages.length) {
							loadStep(stepId+1);
							return;
						}
						imgY = 0;
						fbuf[3] = 0;
						fbuf[4] = styleImages[imgId].width;
						fbuf[5] = styleImages[imgId].height;
		
						simMinCount = Number.MAX_VALUE;
						simMaxCount = 0;
						UnbindTexture(currentShader["img"]);
						BindTexture(currentShader["img"],styleTextures[imgId]);
					}
				}
			}
		}
	}
	setTimeout(gen,0);
}


//load ui for next step if changed
function loadStep(sid) {
	//limit steps to 1-4
	sid = Math.max(0,Math.min(6,sid));
	if (sid === stepId) return;
	stepId = sid;
	
	document.getElementById("set").style.display = stepId===1?"block":"none";
	
	if (stepId === 0) {
		statusText.innerHTML = "Step 1:</br>Select images as source pattern to synthesize from.</br>Your images never leave your device, all processing is done on your GPU.";
		fileSelectEle.style.display = "inline";
		bckButton.style.display = "none";
		
	} else if (stepId === 1) {
		statusText.innerHTML = "Step 2:</br>Fine tune settings.";
		fileSelectEle.style.display = modeSelect.selectedIndex===1?"inline":"none";
		document.getElementById("syo").style.display = modeSelect.selectedIndex===0?"inline":"none";
		bckButton.style.display = nxtButton.style.display = "inline";
		
	} else if (stepId === 2) {
		statusText.innerHTML = "Step 3:</br>Generating...";
		fileSelectEle.style.display = "none";
		nxtButton.style.display = "none";

		//load settings and prepare worker data
		quantization = parseInt(document.getElementById("qt").value,10);
		if (isNaN(quantization)) quantization = 10;
		quantization = Math.max(0,Math.min(100,quantization))/100*2;
		
		patternSz = parseInt(document.getElementById("ps").value,10);
		if (isNaN(patternSz)) patternSz = 10;
		patternSz = Math.max(0,patternSz);
		
		sourceTiling = document.getElementById("stile").checked;
		imageTiling = document.getElementById("itile").checked;
		var tileShader = sourceTiling+imageTiling*2;
		
		//free old transfer textures
		if (transferTextures) {
			for (var i = 0; i < transferTextures.length; i++) gl.deleteTexture(transferTextures[i]);
			transferTextures = null;
		}

		mode = modeSelect.selectedIndex;
		if (mode === 0) {
			var sr = document.getElementById("sr").value.toLowerCase();
			if (sr.indexOf("x") === -1) sr = sr.split(",");
			else sr = sr.split("x");
			if (sr.length === 1) synthWidth = synthHeight = parseInt(sr[0],10);
			else {
				synthWidth = parseInt(sr[0],10);
				synthHeight = parseInt(sr[1],10);
			}
			if (isNaN(synthWidth)) synthWidth = 100;
			synthWidth = Math.max(1,synthWidth);
			if (isNaN(synthHeight)) synthHeight = 100;
			synthHeight = Math.max(1,synthHeight);
			patternSz = Math.min(patternSz,Math.min(synthWidth-1,synthHeight-1));
		} else if (mode === 1) {
			transferTextures = new Array(transferImages.length);
		}
		
		//free old style textures
		if (styleTextures) {
			for (var i = 0; i < styleTextures.length; i++) gl.deleteTexture(styleTextures[i]);
		}
		if (synthTex) {
			gl.deleteTexture(synthTex);
			synthTex = null;
		}
		
		//convert images to textures
		styleTextures = new Array(styleImages.length);
		for (var r = 0, rl = mode<2?mode+1:1; r < rl; r++) {
			var timgs = styleImages,
				dst = styleTextures;
			if (r === 1) {
				timgs = transferImages;
				dst = transferTextures;
			}
			for (var i = 0; i < timgs.length; i++) {
				var tx = gl.createTexture();
				gl.bindTexture(gl.TEXTURE_2D,tx);
				SetTextureOptions(false,false,false,false);
				gl.texImage2D(gl.TEXTURE_2D,0,gl.RGBA,gl.RGBA,gl.UNSIGNED_BYTE,timgs[i]);
				dst[i] = tx;
			}
		}
		
		//free old render textures
		if (styleRT) {
			for (var i = 0; i < styleRT.length; i++) styleRT[i].Delete();
		}
		
		//allocate style render textures, measure max buffer size needed
		styleRT = new Array(styleImages.length);
		var maxBuf = 0;
		for (var i = 0; i < styleImages.length; i++) {
			var si = styleImages[i];
			styleRT[i] = new RenderTexture(si.width,si.height,gl.RGBA,gl.RGBA,gl.UNSIGNED_BYTE);
			maxBuf = Math.max(maxBuf,si.width*si.height);
			patternSz = Math.min(patternSz,Math.min(si.width-1,si.height-1));
		}
		countBuffer = new Int32Array(maxBuf);
		countBytes = new Uint8Array(countBuffer.buffer);
		
		//copy source pixel data to buffers using canvas
		if (mode === 0 || mode === 1) {
			synthSourceData = new Array(styleImages.length);
			var cbe = document.createElement("canvas"),
				c2d = cbe.getContext("2d");
			for (var i = 0; i < styleImages.length; i++) {
				var style = styleImages[i];
				cbe.width = style.width;
				cbe.height = style.height;
				c2d.drawImage(style,0,0);
				synthSourceData[i] = c2d.getImageData(0,0,style.width,style.height);
			}
		}
		
		//initial starting state
		nSeamPasses = styleId = imgId = imgX = imgY = 0;
		if (mode === 0) {//synthesize image mode
			
			//output canvas for synthesized image
			var oce = document.createElement("canvas");
			oce.width = synthWidth;
			oce.height = synthHeight;
			synth2D = oce.getContext("2d");
			synthData = synth2D.getImageData(0,0,1,1);
			imageResults.length = 1;
			imageResults[0] = oce;
			
			//texture to pass synthesized image state to webgl
			synthTex = gl.createTexture();
			gl.bindTexture(gl.TEXTURE_2D,synthTex);
			SetTextureOptions(false,false,false,false);
			gl.texImage2D(gl.TEXTURE_2D,0,gl.RGBA,synthWidth,synthHeight,0,gl.RGBA,gl.UNSIGNED_BYTE,null);
			
			//initial synthesizer uniform and texture state
			shaderPrograms[tileShader].Use();
			BindTexture(currentShader["img"],synthTex);
			
			if (imageTiling) {
				imageSeamPasses = parseInt(document.getElementById("tp").value,10);
				if (isNaN(imageSeamPasses)) imageSeamPasses = 0;
				nSeamPasses = Math.max(1,imageSeamPasses);
				imageSeamPasses = 1;
			}
			
			//img search xy
			fbuf[2] = fbuf[3] = 0;
			//img width/height
			fbuf[4] = synthWidth;
			fbuf[5] = synthHeight;
			fbuf[6] = patternSz;
			fbuf[7] = quantization;
			
		} else if (mode === 1) {//transform existing mode
		
			//output canvas for transformed image
			var oce = document.createElement("canvas"),
				transfer = transferImages[imgId];
			oce.width = transfer.width;
			oce.height = transfer.height;
			synth2D = oce.getContext("2d");
			synthData = synth2D.getImageData(0,0,1,1);
			imageResults.length = 1;
			imageResults[0] = oce;
		
			//initial uniform and texture data
			shaderPrograms[tileShader+4].Use();
			BindTexture(currentShader["img"],transferTextures[imgId]);
			
			//img search xy
			fbuf[2] = fbuf[3] = 0;
			//img width/height
			synthWidth = fbuf[4] = transfer.width;
			synthHeight = fbuf[5] = transfer.height;
			fbuf[6] = patternSz;
			fbuf[7] = quantization;
			
		} else if (mode === 2) {//visualize similarity of source images mode
		
			imageResults.length = 0;
		
			simMinCount = Number.MAX_VALUE;
			simMaxCount = 0;
			
			//similarity shader program initial uniform and texture data
			simBuffer = new Int32Array(maxBuf);
			shaderPrograms[tileShader+4].Use();
			BindTexture(currentShader["img"],styleTextures[imgId]);
			//img search xy
			fbuf[2] = fbuf[3] = 0;
			//img width/height
			fbuf[4] = styleImages[imgId].width;
			fbuf[5] = styleImages[imgId].height;
			fbuf[6] = patternSz;
			fbuf[7] = quantization;
			
		} else alert("Not supported yet");
		
		startTime = Date.now();
		gen();

	} else if (stepId === 3) {
		statusText.innerHTML = "Step 4(Final):</br>Save or copy the images!</br>"+
								"Completed in "+timeStr(Date.now()-startTime);
		bckButton.style.display = "inline";
	}
	
	updateDisplay();
}


//entry point, load app
document.addEventListener("DOMContentLoaded",function() {
	
	//load webgl and default settings
	if (!InitializeQuickWebGL()) {
		alert("Your web browser does not support WebGL.");
		return;
	}
	
	fbuf = new Float32Array(8);
	gl.disable(gl.BLEND);
	gl.disable(gl.DEPTH_TEST);
	gl.enable(gl.CULL_FACE);
	gl.cullFace(gl.BACK);
	
	for (var i = 0; i < 10; i++) {
		var stile = i%2,
			itile = (Math.floor(i/2)%2),
			sim = (i>3?1:0),
			pass = 0;
		if (i > 7) {
			itile = 1;
			sim = 0;
			pass = 1;
		}
		shaderPrograms[i] = new Shader("\n#define SRC_TILING "+stile+
				"\n#define IMG_TILING "+itile+
				"\n#define SIM "+sim+
				"\n#define PASS2 "+pass+
				'\n#include "synth.glsl"',false);
	}

	bgCanvas = document.createElement("canvas");
	bg2d = bgCanvas.getContext("2d");

	//load elements from id
	bckButton = document.getElementById("bck");
	nxtButton = document.getElementById("nxt");
	fileSelectEle = document.getElementById("fs");
	statusText = document.getElementById("st");
	modeSelect = document.getElementById("mo");
	
	//mode select change event
	modeSelect.addEventListener("change",function() {
		fileSelectEle.style.display = modeSelect.selectedIndex===1?"inline":"none";
		document.getElementById("syo").style.display = modeSelect.selectedIndex===0?"inline":"none";
		updateDisplay();
	});

	//image select event
	fileSelectEle.addEventListener("change",function() {
		var files = fileSelectEle.files;
		if (files && files.length) {
			var imgs = new Array(files.length);
			for (var i = 0; i < files.length; i++) {
				var g = document.createElement("img"),
					u = URL.createObjectURL(files[i]);
				g.src = u;
				imgs[i] = g;
			}
			var dst;
			if (stepId === 0) dst = styleImages;
			else dst = transferImages;
			if (dst) {
				for (var i = 0; i < dst.length; i++) URL.revokeObjectURL(dst[i]);
			}
			if (stepId === 0) styleImages = imgs;
			else transferImages = imgs;
			updateDisplay();
		}
	});
	
	//next and back step events
	document.getElementById("bck").addEventListener("click", function() {
		if (stepId >= 2) {
			if (!confirm("Going back will delete the current images, are you sure you want to go back?")) return;
			if (stepId === 3) stepId--;
		}
		loadStep(stepId-1);
	});
	document.getElementById("nxt").addEventListener("click", function() {
		if (!stepComplete()) {
			alert("Please complete the current step before continuing.");
			return;
		}
		loadStep(stepId+1);
	});
	
	//load initial step
	loadStep(0);
});

